/*
 * Grid Widget Implementation
 * Copyright (C) 2000-2001 by Tom Dyas (tdyas@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>
#include <unix_stdarg.h>

#ifdef CONFIG_HANDERA
#include <Vga.h>
#endif
#ifdef CONFIG_SONY
#include <SonyCLIE.h>
 #ifdef Capable
  #undef Capable
 #endif
#endif

#define MAX(x,y) ((x < y)? y: x)
#define TopRowIndex(g) (g->topRow - g->rowCacheOffset)

#include "../db.h"
#include "../grid.h"
#include "../util.h"
#include "../enum.h"
#include "../prefs.h"
#include "../plug.h"
#include "../listview.h"

typedef enum {
    gridPositionNone,
    gridPositionHeaderCell,
    gridPositionHeaderColumnSep,
    gridPositionCell,
    gridPositionColumnSep,
} GridPosition;

static Err _GridInit(GridType* grid, FormPtr form, UInt16 grid_gadget_id,
		    UInt16 grid_checkbox_id, GridModelPtr model,
		    void * model_data) GRIDSECT;


static Err _GridChange(GridType* grid, FormPtr form, GridModelPtr model,
			void * model_data) GRIDSECT;

static void _GridDraw(GridType* grid) GRIDSECT;

static Boolean _GridHandleEvent(GridType* grid, EventPtr event) GRIDSECT;

static void _GridFree(GridType* grid) GRIDSECT;

static Boolean _GridScrollVertical(GridType* grid, WinDirectionType direction,
			       UInt16 increment) GRIDSECT;

static Boolean _GridScrollHorizontal(GridType* grid, WinDirectionType direction,
				 UInt16 increment) GRIDSECT;

static void _GridSetHeaderLook(GridType* grid,
			      Boolean display_header,
			      Boolean display_header_separator, 
			      Boolean header_selectable,
			      FontID header_font) GRIDSECT;

static void _GridSetColumnLook(GridType* grid, Boolean showColumnSep) GRIDSECT;

static UInt16 _GridGetNumberOfRows(GridType* grid) GRIDSECT;

static void _GridHighlightIndex(GridType* grid, UInt16 index) GRIDSECT;

static void _GridUnhighlightAll(GridType* grid) GRIDSECT;

static UInt16 _GridGetTopIndex(const GridType*) GRIDSECT;

static void _GridSetTopIndex(GridType*, UInt16) GRIDSECT;

#ifdef CONFIG_JOGDIAL
static Boolean _GridMoveSelectedLine(GridType* grid, WinDirectionType direction) GRIDSECT;

static void _GridSelectRowIndex(GridType* grid, UInt16 index) GRIDSECT;
#endif
static void _GridGetCellBounds(const GridType*, UInt16, UInt16,
              RectangleType*) GRIDSECT;
static void _GridGetRowBounds(const GridType*, UInt16, RectangleType*) GRIDSECT;
static void _GridGetHeaderBounds(const GridType* grid, UInt16 col, RectangleType* r) GRIDSECT;

static void FillColumnInfo(GridType*) GRIDSECT;
static void InitRowInfo(GridType*) GRIDSECT;
static void InvalidateGrid(GridType*) GRIDSECT;
static void LoadGrid(GridType *) GRIDSECT;
static GridPosition PtToPosition(const GridType*, UInt16 *,
                 UInt16 *, Int16, Int16) GRIDSECT;
static void DrawCellBoolean(const GridType*, RectangleType*, Boolean) GRIDSECT;
static void DrawCell(const GridType*, UInt16, UInt16, void *) GRIDSECT;
static void DrawRow(const GridType*, UInt16, Boolean) GRIDSECT;
#ifdef CONFIG_COLOR
static void ReplaceTwoColors(const GridType*, RectangleType *,
                 UInt16, GridUIColorTableEntries,
                 GridUIColorTableEntries, GridUIColorTableEntries,
                 GridUIColorTableEntries) GRIDSECT;
static void DoCellColorSwap(const GridType*, RectangleType*,
                Boolean, Boolean) GRIDSECT;
#endif
static void SelectCell_Boolean(const GridType*, UInt16, UInt16,
                   EventType* event) GRIDSECT;
static void SelectCell(const GridType*, UInt16, UInt16) GRIDSECT;
static void SelectHeaderCell(const GridType*, UInt16) GRIDSECT;

MemHandle GridColorH;

static void
FillColumnInfo(GridType* grid)
{
    EventType e;
    UInt16 width;
    Int16 x;
    Boolean rightColumnClipped = false;

    x = 0;
    while (x < grid->bounds.extent.x && grid->rightColumn < grid->numCols) {
        /* Ask the grid model for the desired width of this column. */
        width = grid->model->GetColumnWidth(grid->model_data,
                            grid->rightColumn);
        if (FeatureSet.HiRes)
            width = WinScaleCoord(width, true); 
        /* Clip the requested width to the actual space remaining. */
        if (width > grid->bounds.extent.x - x) {
            width = grid->bounds.extent.x - x;
            rightColumnClipped = true;
        }
    
        /* If this is the rightmost column, allocate it all remaining space. */
        if (grid->rightColumn == grid->numCols - 1
            && width < grid->bounds.extent.x - x)
            width = grid->bounds.extent.x - x;
    
        /* Assign the column to this x-offset and allocated width. */
        grid->cols[grid->rightColumn].x = x;
        grid->cols[grid->rightColumn].width = width;
    
        /* Bump the column number up. */
        grid->rightColumn += 1;
    
        /* Move the x-offset ahead by the width of the field, both
         * column sep gutters, and the column sep if necessary.
         */
        x += width + 2 * grid->columnGutterExtent;
        if ((grid->display & Display_Col_sep)) x += 1;
    }

    /* We incremented one too many times. */
    grid->rightColumn -= 1;

    /* Send an event to the form to enable/disable scroll buttons. */
    MemSet(&e, sizeof(e), 0);
    e.eType = gridHorizontalScrollersEvent;
    e.data.generic.datum[0] = grid->controlID;
    e.data.generic.datum[1] = (grid->leftColumn > 0);
    e.data.generic.datum[2] = (grid->rightColumn < grid->numCols - 1)
    || ( rightColumnClipped && (grid->leftColumn < (grid->numCols - 1)) );
    EvtAddEventToQueue(&e);
}

static void
InitRowInfo(GridType* grid)
{
    UInt16 height, i, nextIndex;

    /* Free any existing rows array. (Can happens if look is changed.) */
    if (grid->rows) {
        DeallocateMemPtr(grid->rows);
        grid->rows = 0;
    }

    /* Determine the number of rows and then allocate space for screenRows. */
    height = grid->bounds.extent.y - grid->headerExtent;
    grid->numScreenRows = height / grid->lineHeight;
    grid->numModelRows = grid->model->NumRows(grid->model_data);

	/*
	 * A note about grid->numModelRows
	 *
	 * Make an optimization for small databases.
	 *
	 * Typically, we want to know the number of rows in the model
	 * to ensure that the topRow is set such that the screen is full
	 * this means that many of the calculations involving this value
	 * will underflow if numModelRows < numScreenRows
	 *
	 * For each calculation, we could do MAX(numModelRows,numScreenRows)
	 * but that would be slow.
	 * 
	 * instead we simply do this manipulation.
	 *
	 * The tradeoff is that now, when we increment the model's rows
	 * as a result of adding a record, this value is no longer correct.
	 *
	 * The current way around this is to reload the grid for small 
	 * databases (no real performance hit).  But we could imagine
	 * that we might want an accessor function GridRecordAdded(grid)
	 * and GridRecordRemoved(grid) instead.
	 *
	 */
	grid->numModelRows = MAX(grid->numModelRows,grid->numScreenRows);

#if CONFIG_VLDB
	if ( grid->numModelRows >= CONFIG_VLDB ) {
		grid->numCacheRows = grid->numScreenRows;
		//DbgPrintF( "Experimental Very Large Database Support." );
	}
	else
#endif
		grid->numCacheRows = MAX(grid->numModelRows,grid->numScreenRows);

    grid->rows = AllocateMemPtr(grid->numCacheRows * sizeof(GridScreenRow));
	grid->rowCacheOffset = 0;

#if CONFIG_VLDB	
	ErrFatalDisplayIf( !grid->rows, "Out of Memory." );
#else
	ErrFatalDisplayIf( !grid->rows, "You may need a vldb enabled version" );
#endif

    /* Setup the screenRows array with default values. */
    nextIndex = 0;
    for (i = 0; i < grid->numCacheRows; ++i) {
        grid->rows[i].modelIndex = nextIndex;
        grid->model->Seek(grid->model_data, &(grid ->rows[i].modelIndex), 0, dmSeekForward);
        grid->rows[i].height = grid->lineHeight;
        grid->rows[i].usable = false;
        grid->rows[i].valid = false;
        grid->rows[i].highlighted = false;
        nextIndex = grid->rows[i].modelIndex + 1;
    }

	grid->lastModelIndex = dmMaxRecordIndex;
	grid->model->Seek(grid->model_data, &(grid->lastModelIndex), 0, dmSeekBackward );

}

static Err
_GridChange(GridType* grid, FormPtr form, GridModelPtr model,
     void * model_data)
{
    UInt16 i;

    /* Fill in the model provided by the user. */
    grid->model = model;
    grid->model_data = model_data;
    grid->form = form;
    grid->topRow = 0;

    /* For now, we use the standard font to draw all rows in the grid. */
    grid->lineHeight = FntLineHeight();

    /* Query the model for column informaton and cache in the grid struct. */
    grid->numCols = grid->model->GetColumnCount(grid->model_data);
    if (grid->cols)
        DeallocateMemPtr(grid->cols);
    grid->cols = AllocateMemPtr(grid->numCols * sizeof(GridScreenColumn));
    for (i = 0; i < grid->numCols; ++i) {
        UInt16 width = grid->model->GetColumnWidth(grid->model_data, i);
        if (FeatureSet.HiRes)
            width = WinScaleCoord( width, true);
        grid->cols[i].width = MAX(width, 10);
    }

    /* Now determine what columns are visible and their offsets. */
    grid->leftColumn = grid->rightColumn = 0;

    InitRowInfo(grid);
    FillColumnInfo(grid);
    InvalidateGrid(grid);

    return 0;
}

static Err
_GridInit(GridType* grid, FormPtr form, UInt16 grid_gadget_id,
     UInt16 grid_checkbox_id, GridModelPtr model,
     void * model_data)
{
    Err err;

    if (grid->win){
        _GridFree(grid);
    }

    /* Clear out the entire structure. */
//    MemSet(grid+sizeof(GridOpsType), sizeof(*grid)-sizeof(GridOpsType), 0);

    /* Fill in the model provided by the user. */
    grid->display = 0;
    grid->valid = 0;
    grid->numHighlights = 0;
    grid->controlID = grid_gadget_id;
    grid->ctl_checkbox_id = grid_checkbox_id;
    grid->ctl_checkbox_index = FrmGetObjectIndex(form, grid_checkbox_id);
    grid->ctl_checkbox = FrmGetObjectPtr(form, grid->ctl_checkbox_index);

    /* Use the form to pull the bounds out of the gadget. */
    FrmGetObjectBounds(form, FrmGetObjectIndex(form, grid_gadget_id),
               &(grid->bounds));

#ifdef CONFIG_SONY
    if (FeatureSet.Sony && SonyLibRefNum) {
        grid->bounds.extent.x *= 2;
        grid->bounds.extent.y *= 2;
        grid->bounds.topLeft.x *= 2;
        grid->bounds.topLeft.y *= 2;
        grid->win = HRWinCreateOffscreenWindow(SonyLibRefNum,
                     grid->bounds.extent.x,
                     grid->bounds.extent.y,
                     screenFormat, &err);
    } else
#endif
    if (FeatureSet.HiRes) {
        WinScaleRectangle(&(grid->bounds));
    /* Allocate an offscreen window for flicker-free drawing. */
        grid->win = WinCreateOffscreenWindow(grid->bounds.extent.x,
                     grid->bounds.extent.y,
                     nativeFormat, &err);
    } else {
        grid->win = WinCreateOffscreenWindow(grid->bounds.extent.x,
                     grid->bounds.extent.y,
                     screenFormat, &err);
    }

    /* Bail out now if we could not allocate the offscreen window. */
    if (!grid->win) {
        return err;
    }

    /* Initialize the color scheme that will be used by the grid. */
    grid->colors = NULL;
#ifdef CONFIG_COLOR
    if (FeatureSet.OS35) {
        UInt16 size;
        size = MemHandleSize(GridColorH) - 2 * sizeof(UInt16);
        grid->colors = AllocateMemPtr(size);
        MemMove(grid->colors, LockMemHandle(GridColorH) + 2 * sizeof(UInt16), size);
        UnlockMemHandle(GridColorH);
        //grid->colors = LockMemHandle(GridColorH) + 2 * sizeof(UInt16) ;
    }
#endif

    grid->display = DBPrefs.GridDisplay;
    _GridSetHeaderLook(grid, DBPrefs.GridDisplay & Display_Header,
                            DBPrefs.GridDisplay & Display_Header_sep,
                            true,
                            DBPrefs.TheBoldFont);
    _GridSetColumnLook(grid, DBPrefs.GridDisplay & Display_Col_sep);

    /* All the UI components are initially invalid. */
    grid->valid = 0;

    err = _GridChange( grid, form, model, model_data);

    /* Fill in the remainder of the fields. */
    grid->topRow = 0;

    grid->screenRowSelected = NoneScreenRowSelected;

    return err;
}


static void
_GridFree(GridType* grid)
{

    if (grid->rows) {
        DeallocateMemPtr(grid->rows);
        grid->rows = 0;
    }
    if (grid->cols) {
        DeallocateMemPtr(grid->cols);
        grid->cols = 0;
    }
    if (grid->win) {
        WinDeleteWindow(grid->win, false);
        grid->win = 0;
    }

    if (grid->colors) {
        DeallocateMemPtr(grid->colors);
        grid->colors = NULL;
//        UnlockMemHandle(GridColorH);
    }

}

/* Invalidate the entire grid so that GridDraw() will redraw it all. */
static void
InvalidateGrid(GridType* grid)
{
    UInt16 i;

    if (grid->topRow > grid->numModelRows - grid->numScreenRows)
        grid->topRow = grid->numModelRows - grid->numScreenRows;

    /* Set all state variables to the invalid position. */
    grid->valid = 0;
    for (i = 0; i < grid->numScreenRows && (TopRowIndex(grid) + i) < grid->numCacheRows; ++i) {
        grid->rows[TopRowIndex(grid) + i].valid = false;
    }
}


/* Figure out what rows to display in the grid. */
static void
LoadGrid(GridType* grid)
{
    UInt16 row, index, end;
    Boolean scrollUp, scrollDown;
    EventType e;

    /* Adjust the top visible row so it is valid. */
    grid->model->Seek(grid->model_data, &(grid->rows[TopRowIndex(grid)].modelIndex), 0, dmSeekForward);

    /* Make sure that the entire display area is filled if possible. */
    if ( grid->numModelRows - grid->numScreenRows > 0) {
        if ( grid->topRow + grid->numScreenRows > grid->numModelRows)
            grid->topRow = grid->numModelRows - grid->numScreenRows;
    } else
        grid->topRow =  0;

    index = grid->rows[TopRowIndex(grid)].modelIndex;
	end = TopRowIndex(grid) + grid->numScreenRows;
    for (row = TopRowIndex(grid); row < grid->numCacheRows && row < end; ++row) {
        /* Stop the loop if no more rows are available from the model. */
        if (! grid->model->Seek(grid->model_data, &index, 0, dmSeekForward))
            break;

        /* Mark this row for drawing if it had previously held no
         * index at all or another index from the model.
         */
        if (! grid->rows[row].usable) {
            grid->rows[row].usable = true;
            grid->rows[row].valid = false;
            grid->rows[row].modelIndex = index;
            grid->rows[row].height = grid->lineHeight;
        }

        /* Increment the row index past the current index. */
        if (! grid->model->Seek(grid->model_data, &index, 1, dmSeekForward)) {
            row ++;
            break;
        }
    }

    /* Mark the remaining rows as unusable. */
    while (row < grid->numCacheRows && row < end) {
        grid->rows[row].usable = false;
        grid->rows[row].valid = false;
        ++row;
    }

    if (DBPrefs.TheBoldFont != grid->header_font){
        _GridSetHeaderLook(grid, DBPrefs.GridDisplay & Display_Header,
                                DBPrefs.GridDisplay & Display_Header_sep,
                                true, DBPrefs.TheBoldFont);
    }

    // Determine if the model can be scrolled up and down. 
    scrollUp = false;
    if (grid->topRow > 0) {
        scrollUp = true;
    }
    scrollDown = false;
    if (grid->topRow < grid->numModelRows - grid->numScreenRows) {
        scrollDown = true;
    }

    // Send a notification event to the form about the vertical scrollers.
    MemSet(&e, sizeof(e), 0);
    e.eType = gridVerticalScrollersEvent;
    e.data.generic.datum[0] = grid->controlID;
    e.data.generic.datum[1] = scrollUp;
    e.data.generic.datum[2] = scrollDown;
    EvtAddEventToQueue(&e);
}

static void
_GridGetCellBounds(const GridType* grid, UInt16 screenRow, UInt16 col,
          RectangleType* r)
{
    UInt16 i;

    if (grid->topRow + screenRow > grid->numModelRows) {
        r->extent.y = grid->lineHeight;
        r->topLeft.y = grid->headerExtent;
        for (i = 0; screenRow > 0; ++i, --screenRow) {
            r->topLeft.y += grid->lineHeight;
        }
    } else {        
        /* Determine the top of the cell. */
        r->extent.y = grid->rows[TopRowIndex(grid) + screenRow].height;
        r->topLeft.y = grid->headerExtent;
        for (i = 0; screenRow > 0; ++i, --screenRow) {
            r->topLeft.y += grid->rows[TopRowIndex(grid) + i].height;
        }
    }

    /* Determine the left of the cell. */
    r->topLeft.x = grid->cols[col].x;
    r->extent.x = grid->cols[col].width;
}

static void
_GridGetRowBounds(const GridType* grid, UInt16 screenRow, RectangleType* r)
{
    UInt16 i;

    if (grid->topRow + screenRow > grid->numModelRows) {
        r->extent.y = grid->lineHeight;
        r->topLeft.y = grid->headerExtent;
        for (i = 0; screenRow > 0; ++i, --screenRow) {
            r->topLeft.y += grid->lineHeight;
        }

        r->topLeft.x = 0;
        r->extent.x = grid->bounds.extent.x;
        return;
    }
        
    /* Determine the top of the cell. */
    r->extent.y = grid->rows[TopRowIndex(grid) + screenRow].height;
    r->topLeft.y = grid->headerExtent;
	if ( screenRow < grid->numCacheRows ) {
		for (i = 0; screenRow > 0; ++i, --screenRow) {
			r->topLeft.y += grid->rows[TopRowIndex(grid) + i].height;
		}
	}
	else {
		// push offscreen 
		r->topLeft.y += grid->numCacheRows * grid->lineHeight;
	}

    r->topLeft.x = 0;
    r->extent.x = grid->bounds.extent.x;
}

static void
_GridGetHeaderBounds(const GridType* grid, UInt16 col, RectangleType* r)
{
    /* Determine the top of the cell. */
    r->extent.y = grid->headerExtent;
    r->topLeft.y = grid->bounds.topLeft.y;
    r->topLeft.x = grid->cols[col].x;
    r->extent.x = grid->cols[col].width;
}

static GridPosition
PtToPosition(const GridType* grid, UInt16 * screenRow, UInt16 * col,
         Int16 x, Int16 y)
{
    /* Normalize the point into local coordinates. */
    x -= grid->bounds.topLeft.x;
    y -= grid->bounds.topLeft.y;

    /* Determine if the header row was hit. */
    if (grid->display & Display_Header) {
        if (y < grid->headerExtent) {
            *col = grid->leftColumn;
            while (*col < grid->numCols && x > grid->cols[*col].width + 2*grid->columnGutterExtent
                                       + ((grid->display & Display_Col_sep) ? 1 : 0)) {
                x -= grid->cols[*col].width - 2*grid->columnGutterExtent
                                - ((grid->display & Display_Col_sep) ? 1 : 0);
                *col += 1;
            }
    
            if (x < grid->cols[*col].width)
                return gridPositionHeaderCell;
            else if (x >= grid->cols[*col].width
                         && x < (grid->cols[*col].width
                             + 2*grid->columnGutterExtent
                             + ((grid->display & Display_Col_sep) ? 1 : 0)))
                return gridPositionHeaderColumnSep;
        }
    
        /* Remove the header row from the y coordinate. */
        y -= grid->headerExtent;
    }

    /* Determine the row where the point lies. */
    *screenRow = 0;
    while (*screenRow < grid->numScreenRows 
            && y >= grid->rows[TopRowIndex(grid) + *screenRow].height) {
        y -= grid->rows[TopRowIndex(grid) + *screenRow].height;
        *screenRow += 1;
    }

    /* Determine the column where the point lies. */
    *col = grid->leftColumn;
    while (*col < grid->numCols && x >= (grid->cols[*col].width + 2*grid->columnGutterExtent
                               + ((grid->display & Display_Col_sep) ? 1 : 0))) {
        x -= (grid->cols[*col].width + 2*grid->columnGutterExtent
            + ((grid->display & Display_Col_sep) ? 1 : 0));
        *col += 1;
    }
    if (x < grid->cols[*col].width)
        return gridPositionCell;
    else if (x >= grid->cols[*col].width
                            && x < (grid->cols[*col].width + 2*grid->columnGutterExtent
                                + ((grid->display & Display_Col_sep) ? 1 : 0)))
        return gridPositionColumnSep;
    else
        return gridPositionNone;
}

#if 1
static void
DrawCellBoolean(const GridType* grid, RectangleType* r, Boolean value)
{

    /* Clear out any extra stuff from the cell. */
    WinEraseRectangle(r, 0);

    /* Setup the checkbox structure for this cell. */
    CtlSetEnabled(grid->ctl_checkbox, 1);
    CtlSetUsable(grid->ctl_checkbox, true);

    if (FeatureSet.HiRes) {
        WinUnscaleRectangle(r);
    }

    FrmSetObjectBounds(grid->form, grid->ctl_checkbox_index, r);
    CtlSetValue(grid->ctl_checkbox, value);

    /* Now instruct the checkbox to draw itself. */
    CtlShowControl(grid->ctl_checkbox);

    /* Make the sure the checkbox is not usable again. */
    CtlSetEnabled(grid->ctl_checkbox, 0);
    CtlSetUsable(grid->ctl_checkbox, 0);
}
#else
static void
DrawCellBoolean(const GridType* grid, RectangleType* r, Boolean value)
{
	FontID oldfont;

    WinEraseRectangle(r, 0);
	oldfont = FntSetFont(4);
	if (value)
		WinDrawChar(0x01, r->topLeft.x, r->topLeft.y);
	else
		WinDrawChar(0x00, r->topLeft.x, r->topLeft.y);
	FntSetFont(oldfont);
}
#endif

static void
DrawCell(const GridType* grid, UInt16 row, UInt16 col, void * data)
{
    RectangleType r;
    Char buf[128];
    Boolean b;
	GridCellType type;

    /* Retrieve the location of this cell. */
    _GridGetCellBounds(grid, row - TopRowIndex(grid), col, &r);
    //there is a trouble with the bound
    if (!r.extent.x || !r.extent.y)
        return;

	type = (grid->model->GetCellType(grid->model_data,
					 grid->rows[row].modelIndex, col));
    switch ( type ) {
	case gridCellTypeRJString:								  
    case gridCellTypeNote:
    case gridCellTypeString:
      grid->model->GetCellString(grid->model_data,
				 grid->rows[row].modelIndex,
				 col, buf, sizeof(buf), data);
      buf[sizeof(buf)-1] = '\0';
      
      WinEraseRectangle(&r, 0);
      UtilWinDrawTruncChars(buf, StrLen(buf),
			    r.topLeft.x, r.topLeft.y, r.extent.x, 
			    (type == gridCellTypeRJString) ? 0:1);
      break;
      
    case gridCellTypeBoolean:
        b = grid->model->GetCellBoolean(grid->model_data,
                        grid->rows[row].modelIndex,
                        col, data);
        DrawCellBoolean(grid, &r, b);
    break;
    }
}

static void
DrawRow(const GridType* grid, UInt16 row, Boolean highlighted)
{
    GridScreenRow* screenRow = &grid->rows[row];
    void * data;
    UInt16 col;


    data = grid->model->LockIndex(grid->model_data,
                  screenRow->modelIndex, false);


#ifdef CONFIG_COLOR
    if (FeatureSet.OS35) {
        if (!highlighted && (grid->screenRowSelected - grid->rowCacheOffset != row)) {
            WinSetBackColor(grid->colors[(row % 2)?GridUICellBackground:GridUICellBackground2]);
            WinSetTextColor(grid->colors[GridUICellForeground]);
        } else {
            WinSetBackColor(grid->colors[GridUICellHighlightedBackground]);
            WinSetTextColor(grid->colors[GridUICellHighlightedForeground]);
        }
    }
#endif
    for (col = grid->leftColumn; col <= grid->rightColumn; ++col) {
        DrawCell(grid, row, col, data);
    }

    grid->model->UnlockIndex(grid->model_data, screenRow->modelIndex,
                 false, data);
}

static void
_GridDraw(GridType* grid)
{
    UInt16 col, row;
    RectangleType r;
    Char buf[128];
    WinHandle winH;
    FontID oldFont;


    /* On PalmOS 3.5+, save the current draw state. */
    if (FeatureSet.OS35) {
        WinPushDrawState();
    }

  
    /* Update the usable and valid flags of the displayable rows. */
    LoadGrid(grid);

    /* Switch the draw window to the offscreen window. */
    winH = WinGetDrawWindow();
    WinSetDrawWindow(grid->win);

    /* Draw the column headers and column markers. */
    if ((grid->display & Display_Header) && !(grid->valid & Valid_Header)) {
        /* Switch to the header's color scheme. */
#ifdef CONFIG_COLOR
        if (FeatureSet.OS35) {
            WinSetBackColor(grid->colors[GridUIHeaderBackground]);
            WinSetTextColor(grid->colors[GridUIHeaderText]);
        }
#endif
    
        /* Clear the header to the background color. */
        r.topLeft.x = r.topLeft.y = 0;
        r.extent.x  = grid->bounds.extent.x;
        r.extent.y  = grid->headerExtent;
        WinEraseRectangle(&r, 0);

        /* Render the name of each column. */
        for (col = grid->leftColumn; col <= grid->rightColumn; ++col) {
            /* Retrieve the column name. */
            grid->model->GetColumnName(grid->model_data, col, buf, sizeof buf);
            buf[sizeof(buf) - 1] = '\0';
    
            /* Draw the column name. */
            oldFont = FntSetFont(grid->header_font);

            UtilWinDrawTruncChars(buf, StrLen(buf), grid->cols[col].x, 0,
				                        grid->cols[col].width,
				                        (grid->model->GetCellType(grid->model_data,
								        0, col) != gridCellTypeRJString));
            FntSetFont(oldFont);
        }

#ifdef CONFIG_COLOR
        if (FeatureSet.OS35) {
            WinSetBackColor(grid->colors[GridUICellBackground]);
            WinSetTextColor(grid->colors[GridUICellForeground]);
        }
#endif
        /* Mark the header valid now that it has been drawn. */
        grid->valid |= Valid_Header;

        /* Draw the seperator between the header and data rows. */
        if ((grid->display & Display_Header_sep) && !(grid->valid & Valid_Header_sep)) {
#ifdef CONFIG_COLOR
            if (FeatureSet.OS35) {
                WinSetForeColor(grid->colors[GridUIHeaderSeparator]);
            }
#endif
            WinDrawLine(0, grid->headerExtent - 1,
                grid->bounds.extent.x, grid->headerExtent - 1);

#ifdef CONFIG_COLOR
            if (FeatureSet.OS35) {
                WinSetBackColor(grid->colors[GridUICellBackground]);
            }
#endif
            grid->valid |= Valid_Header_sep;
        }
    }

   /* Scan through each row and see what cells need to be redrawn. */
    for (row = TopRowIndex(grid); row < TopRowIndex(grid) + grid->numScreenRows; ++row) {

        if (row < grid->numCacheRows && grid->rows[row].usable) {
            if (! grid->rows[row].valid) {
                DrawRow(grid, row, grid->rows[row].highlighted);
                grid->rows[row].valid = true;
            }
        } else if ( row < grid->numCacheRows && ! grid->rows[row].valid) {
            _GridGetRowBounds(grid, row - TopRowIndex(grid), &r);
#ifdef CONFIG_COLOR
            if (FeatureSet.OS35) {
                WinSetBackColor(grid->colors[(row % 2)?GridUICellBackground:GridUICellBackground2]);
            }
#endif
            WinEraseRectangle(&r, 0);
            grid->rows[row].valid = true;
        }
    }

    /* Draw the vertical cell seperator lines. */
    if ((grid->display & Display_Col_sep) && !(grid->valid & Valid_Col_sep)) {
#ifdef CONFIG_COLOR
        if (FeatureSet.OS35) {
            WinSetForeColor(grid->colors[GridUICellColumnSeparator]);
        }
#endif
        for (col = grid->leftColumn + 1; col <= grid->rightColumn; ++col) {
            r.topLeft.x = grid->cols[col].x - 2 * grid->columnGutterExtent - 1;
            r.topLeft.y = 0;
            r.extent.x = 3;
            r.extent.y = grid->bounds.extent.y;
            WinEraseRectangle(&r, 0);
            WinDrawLine(grid->cols[col].x - grid->columnGutterExtent - 1,
                0,
                grid->cols[col].x - grid->columnGutterExtent - 1,
                grid->bounds.extent.y);
        }
#ifdef CONFIG_COLOR
        if (FeatureSet.OS35) {
            WinSetBackColor(grid->colors[GridUICellBackground]);
        }
#endif
        grid->valid |= Valid_Col_sep;
    }

    /* Erase any remaining space in the grid not used by a row. */
    r.topLeft.x = 0;
    r.topLeft.y = grid->headerExtent + grid->numScreenRows * grid->lineHeight;
    r.extent.x = grid->bounds.extent.x;
    r.extent.y = grid->bounds.extent.y - r.topLeft.y;
#ifdef CONFIG_COLOR
    if (FeatureSet.OS35) {
        WinSetBackColor(grid->colors[GridUICellBackground]);
    }
#endif
    WinEraseRectangle(&r, 0);

    /* Reset back to the original draw window. */
    WinSetDrawWindow(winH);

    /* Copy the offscreen window to the actual screen. */
    r.topLeft.x = 0;
    r.topLeft.y = 0;
    r.extent.x = grid->bounds.extent.x;
    r.extent.y = grid->bounds.extent.y;
#ifdef CONFIG_SONY
	if (FeatureSet.Sony && SonyLibRefNum) {
		HRWinCopyRectangle(SonyLibRefNum, grid->win, 0, &r, grid->bounds.topLeft.x,
             grid->bounds.topLeft.y, winPaint);
	} else
#endif
    WinCopyRectangle(grid->win, 0, &r, grid->bounds.topLeft.x,
             grid->bounds.topLeft.y, winPaint);

    /* Restore the original draw state. */
    if (FeatureSet.OS35) {
        WinPopDrawState();
    }
}

/* Swap colors on the screen. Can only be called on PalmOS 3.5 or higher. */
#ifdef CONFIG_COLOR
static void
ReplaceTwoColors(const GridType* grid,
         RectangleType * rP, UInt16 cornerDiam,
         GridUIColorTableEntries oldForeground,
         GridUIColorTableEntries oldBackground,
         GridUIColorTableEntries newForeground,
         GridUIColorTableEntries newBackground)
{
    IndexedColorType oldForegroundIndex = grid->colors[oldForeground];
    IndexedColorType oldBackgroundIndex = grid->colors[oldBackground];
    IndexedColorType newForegroundIndex = grid->colors[newForeground];
    IndexedColorType newBackgroundIndex = grid->colors[newBackground];

    WinPushDrawState();
    WinSetDrawMode(winSwap);
    WinSetPatternType(blackPattern);

    if (newBackgroundIndex == oldForegroundIndex) {
        if (newForegroundIndex == oldBackgroundIndex) {
            /* Handle the case when foreground and background colors
             * change places, such as on black and white systems, with
             * a single swap.
             */
            WinSetBackColor(oldBackgroundIndex);
            WinSetForeColor(oldForegroundIndex);
#ifdef CONFIG_SONY
            if (FeatureSet.Sony && SonyLibRefNum) {
                HRWinPaintRectangle(SonyLibRefNum, rP, cornerDiam);
            } else
#endif
            WinPaintRectangle(rP, cornerDiam);
        } else {
            /* Handle the case when the old foreground and the new
             * background are the same, using two swaps.
             */
            WinSetBackColor(oldForegroundIndex);
            WinSetForeColor(oldBackgroundIndex);
#ifdef CONFIG_SONY
            if (FeatureSet.Sony && SonyLibRefNum) {
                HRWinPaintRectangle(SonyLibRefNum, rP, cornerDiam);
            } else
#endif
            WinPaintRectangle(rP, cornerDiam);
            WinSetBackColor(oldBackgroundIndex);
            WinSetForeColor(newForegroundIndex);
#ifdef CONFIG_SONY
            if (FeatureSet.Sony && SonyLibRefNum) {
                HRWinPaintRectangle(SonyLibRefNum, rP, cornerDiam);
            } else
#endif
            WinPaintRectangle(rP, cornerDiam);
        }
    } else if (oldBackgroundIndex == newForegroundIndex) {
        /* Handle the case when the old background and the new
         * foreground are the same, using two swaps.
         */
        WinSetBackColor(newForegroundIndex);
        WinSetForeColor(oldForegroundIndex);
#ifdef CONFIG_SONY
		if (FeatureSet.Sony && SonyLibRefNum) {
			HRWinPaintRectangle(SonyLibRefNum, rP, cornerDiam);
		} else
#endif
        WinPaintRectangle(rP, cornerDiam);
        WinSetBackColor(newBackgroundIndex);
        WinSetForeColor(oldForegroundIndex);
#ifdef CONFIG_SONY
        if (FeatureSet.Sony && SonyLibRefNum) {
            HRWinPaintRectangle(SonyLibRefNum, rP, cornerDiam);
        } else
#endif
        WinPaintRectangle(rP, cornerDiam);
    } else  {
        /* Handle the case when no two colors are the same, as is
         * typically the case on color systems, using two swaps.
         */
        WinSetBackColor(oldBackgroundIndex);
        WinSetForeColor(newBackgroundIndex);
#ifdef CONFIG_SONY
		if (FeatureSet.Sony && SonyLibRefNum) {
			HRWinPaintRectangle(SonyLibRefNum, rP, cornerDiam);
		} else
#endif
        WinPaintRectangle(rP, cornerDiam);
        WinSetBackColor(oldForegroundIndex);
        WinSetForeColor(newForegroundIndex);
#ifdef CONFIG_SONY
        if (FeatureSet.Sony && SonyLibRefNum) {
            HRWinPaintRectangle(SonyLibRefNum, rP, cornerDiam);
        } else
#endif
        WinPaintRectangle(rP, cornerDiam);
    }

    WinPopDrawState();
}
#endif

static void
SelectCell_Boolean(const GridType* grid, UInt16 screenRow, UInt16 col,
           EventType* event)
{
    RectangleType screen_rect, offscreen_rect;
    Boolean value, orig_value;
    WinHandle winH;
    void * row_data;
    EventType new_event;

    /* Save the current draw state. */
    if (FeatureSet.OS35) {
        WinPushDrawState();
    }
  
    /* Determine the location of this cell on-screen and off-screen. */
    _GridGetCellBounds(grid, screenRow, col, &offscreen_rect);
    RctCopyRectangle(&offscreen_rect, &screen_rect);
    screen_rect.topLeft.x += grid->bounds.topLeft.x;
    screen_rect.topLeft.y += grid->bounds.topLeft.y;

    /* Lock down this index in the model. */
    row_data = grid->model->LockIndex(grid->model_data,
                      grid->rows[TopRowIndex(grid) + screenRow].modelIndex, true);

    /* Query the model for the current value of the cell. */
    value = grid->model->GetCellBoolean(grid->model_data,
                    grid->rows[TopRowIndex(grid) + screenRow].modelIndex,
                    col, row_data);
    orig_value = value;

    /* Setup the checkbox structure for this cell. */
    CtlSetEnabled(grid->ctl_checkbox, 1);
    CtlSetUsable(grid->ctl_checkbox, 1);
    FrmSetObjectBounds(grid->form, grid->ctl_checkbox_index, &screen_rect);
    CtlSetValue(grid->ctl_checkbox, value);
    CtlDrawControl(grid->ctl_checkbox);

    /* Fake a ctlEnterEvent for the checkbox control. */
    EvtCopyEvent(event, &new_event);
    new_event.eType = ctlEnterEvent;
    new_event.data.ctlEnter.controlID = grid->ctl_checkbox_id;
    new_event.data.ctlEnter.pControl = grid->ctl_checkbox;

    while (true) {
        if (new_event.eType == ctlEnterEvent) {
            CtlHandleEvent(grid->ctl_checkbox, &new_event);
        } else if (new_event.eType == ctlSelectEvent) {
            value = CtlGetValue(grid->ctl_checkbox);
            /* SendTableSelectEvent (table, row, column); */
        } else if (new_event.eType == ctlExitEvent) {
            /* Post a tableExitEvent. */
            /* SendTableExitEvent (table, row, column); */
        } else {
            /* If we got an event that we didn't expect, put it back
             * in the queue.
             */
            EvtAddEventToQueue(&new_event);
            break;
        }
    
        /* Retrieve the next event. */
        EvtGetEvent(&new_event, evtWaitForever);
    }

    /* Disable the checkbox so it not handled by the form. */
    CtlSetEnabled(grid->ctl_checkbox, 0);
    CtlSetUsable(grid->ctl_checkbox, 0);

    /* If the user completed the selection, then inform the model. */
    if (value != orig_value) {
        grid->model->SetCellBoolean(grid->model_data,
                        grid->rows[TopRowIndex(grid) + screenRow].modelIndex,
                        col, value, row_data);
    
        winH = WinGetDrawWindow();
        WinSetDrawWindow(grid->win);
        DrawCellBoolean(grid, &offscreen_rect, value);
        WinSetDrawWindow(winH);
    }

    /* Unlock this index in the model. */
    grid->model->UnlockIndex(grid->model_data,
                 grid->rows[TopRowIndex(grid) + screenRow].modelIndex,
                 true, row_data);

    /* Restore the previous draw state. */
    if (FeatureSet.OS35) {
        WinPopDrawState();
    }
}

#ifdef CONFIG_COLOR
static void
DoCellColorSwap(const GridType* grid, RectangleType* r,
        Boolean highlighted, Boolean selected)
{
    GridUIColorTableEntries old_fore, old_back, new_fore, new_back, tmp;

    if (highlighted) {
        old_fore = GridUICellHighlightedForeground;
        old_back = GridUICellHighlightedBackground;
        new_fore = GridUICellSelectedForeground;
        new_back = GridUICellSelectedBackground;
    } else {
        old_fore = GridUICellForeground;
        old_back = GridUICellBackground;
        new_fore = GridUICellSelectedForeground;
        new_back = GridUICellSelectedBackground;
    }

    if (!selected) {
#define swap_color(a,b) tmp = a; a = b; b = tmp
        swap_color(old_fore, new_fore);
        swap_color(old_back, new_back);
#undef swap_color
    }

    ReplaceTwoColors(grid, r, 0, old_fore, old_back, new_fore, new_back);
}
#endif

static void
SelectCell(const GridType* grid, UInt16 screenRow, UInt16 col)
{
    RectangleType offscreen_rect, screen_rect;
    Boolean selected, highlighted;
    Coord x, y;

    /* Save the current draw state. */
    if (FeatureSet.OS35) {
        WinPushDrawState();
    }

    /* Retrieve the location of the row that was tapped. */
    _GridGetRowBounds(grid, screenRow, &offscreen_rect);
    RctCopyRectangle(&offscreen_rect, &screen_rect);
    screen_rect.topLeft.x += grid->bounds.topLeft.x;
    screen_rect.topLeft.y += grid->bounds.topLeft.y;

    /* Draw the row in the highlighted state and put it on the screen. */
    selected = true;
    highlighted = grid->rows[TopRowIndex(grid) + screenRow].highlighted;
#ifdef CONFIG_COLOR
    if (FeatureSet.OS35) {
        DoCellColorSwap(grid, &screen_rect, highlighted, selected);
    } else
#endif
    WinInvertRectangle(&screen_rect, 0);
    /* Loop until the user lifts the pen. */
    while (true) {
        Boolean redraw, penDown;
    
        /* Get the current pen status. Leave the loop when it is lifted. */
#ifdef CONFIG_SONY
	    if( FeatureSet.Sony && SonyLibRefNum) {
            PenGetPoint(&x, &y, &penDown); //PenGetPoint = EvtGetPen
			x *= 2;
			y *= 2;
		}
        else
#endif
            PenGetPoint(&x, &y, &penDown);//PenGetPoint = EvtGetPen
        if (!penDown)
            break;

        WinDisplayToWindowPt(&x, &y);

        /* Check to see if the pen moved in or out of the row. */
        redraw = false;
        if (RctPtInRectangle(x, y, &screen_rect)) {
            if (! selected) {
				selected = true;
				redraw = true;
            }           
        } else {
            if (selected) {
				selected = false;
				redraw = true;
            }
        }
    
        /* If the state of the selection changed, redraw the row. */
        if (redraw) {
#ifdef CONFIG_COLOR
            if (FeatureSet.OS35) {
                DoCellColorSwap(grid, &screen_rect, highlighted, false);
            } else
#endif
            WinInvertRectangle(&screen_rect, 0);
        }
    }
        
    /* Uninvert the selection if the pen is still over it. */
    if (selected) {
        EventType e;
    
        /* Restore the unselected appearance of the row. */
#ifdef CONFIG_COLOR
        if (FeatureSet.OS35) {
            DoCellColorSwap(grid, &screen_rect, highlighted, false);
        } else
#endif
        WinInvertRectangle(&screen_rect, 0);

        /* Inform the system about the selection. */
        MemSet(&e, sizeof(e), 0);
        e.eType = gridSelectEvent;
        e.data.tblSelect.tableID = grid->controlID;
        e.data.tblSelect.pTable = (TableType *) grid;
        e.data.tblSelect.row = grid->rows[TopRowIndex(grid) + screenRow].modelIndex;
        e.data.tblSelect.column = col;
        EvtAddEventToQueue(&e);
    }

    /* Restore the previous draw state. */
    if (FeatureSet.OS35) {
        WinPopDrawState();
    }
}

static void
SelectHeaderCell(const GridType* grid, UInt16 col)
{
    RectangleType offscreen_rect, screen_rect;
    Boolean selected;
    Coord x, y;
    char buf[128];
    FontID oldFont;

    /* Save the current draw state and switch to header font. */
    if (FeatureSet.OS35) {
        WinPushDrawState();
    }

    oldFont = FntSetFont(grid->header_font);

    /* Construct rectangles representing the location of the header cell. */
    offscreen_rect.topLeft.x = grid->cols[col].x;
    offscreen_rect.topLeft.y = 0;
    offscreen_rect.extent.x = grid->cols[col].width;
    offscreen_rect.extent.y = grid->headerExtent;
    RctCopyRectangle(&offscreen_rect, &screen_rect);
    screen_rect.topLeft.x += grid->bounds.topLeft.x;
    screen_rect.topLeft.y += grid->bounds.topLeft.y;

    /* Retrieve the header text. */
    MemSet(buf, sizeof(buf), 0);
    grid->model->GetColumnName(grid->model_data, col, buf, sizeof(buf));
    buf[sizeof(buf) - 1] = '\0';

    /* Draw the header cell in the highlighted state. */
    selected = true;
#ifdef CONFIG_COLOR
    if (FeatureSet.OS35) {
        ReplaceTwoColors(grid, &screen_rect, 0,
                 GridUIHeaderText,
                 GridUIHeaderBackground,
                 GridUIHeaderSelectedText,
                 GridUIHeaderSelectedBackground);
    } else
#endif
    WinInvertRectangle(&screen_rect, 0);

    /* Loop until the user lifts the pen. */
    while (true) {
        Boolean redraw, penDown;

        /* Get the current pen status. Leave the loop when it is lifted. */
#ifdef CONFIG_SONY
	    if( FeatureSet.Sony && SonyLibRefNum) {
            PenGetPoint(&x, &y, &penDown); //PenGetPoint = EvtGetPen
			x *= 2;
			y *= 2;
		}
        else
#endif
            PenGetPoint(&x, &y, &penDown); //PenGetPoint = EvtGetPen
        if (!penDown)
            break;

        WinDisplayToWindowPt(&x, &y);

        /* Check to see if the pen moved in or out of the header cell. */
        redraw = false;
        if (RctPtInRectangle(x, y, &screen_rect)) {
            if (! selected) {
                selected = true;
                redraw = true;
            }
        } else {
            if (selected) {
                selected = false;
                redraw = true;
            }
        }

        /* If the state of the selection changed, redraw the row. */
        if (redraw) {
#ifdef CONFIG_COLOR
            if (FeatureSet.OS35) {
                if (selected) {
                    ReplaceTwoColors(grid, &screen_rect, 0,
                             GridUIHeaderText,
                             GridUIHeaderBackground,
                             GridUIHeaderSelectedText,
                             GridUIHeaderSelectedBackground);
                } else {
                    ReplaceTwoColors(grid, &screen_rect, 0,
                             GridUIHeaderSelectedText,
                             GridUIHeaderSelectedBackground,
                             GridUIHeaderText,
                             GridUIHeaderBackground);
                }
            } else
#endif
            WinInvertRectangle(&screen_rect, 0);
       }
    }

    /* Uninvert the selection if the pen is still over it. */
    if (selected) {
        EventType e;

        /* Restore the unselected appearance of the row. */
#ifdef CONFIG_COLOR
        if (FeatureSet.OS35) {
            ReplaceTwoColors(grid, &screen_rect, 0,
                     GridUIHeaderSelectedText,
                     GridUIHeaderSelectedBackground,
                     GridUIHeaderText,
                     GridUIHeaderBackground);
        } else
#endif
        WinInvertRectangle(&screen_rect, 0);

        /* Inform the system about the selection. */
        MemSet(&e, sizeof(e), 0);
        e.eType = gridHeaderSelectEvent;
        e.data.tblSelect.tableID = grid->controlID;
        e.data.tblSelect.pTable = (TableType *) grid;
        e.data.tblSelect.row = 0;
        e.data.tblSelect.column = col;
        EvtAddEventToQueue(&e);
    }

    /* Restore the previous draw state. */
    FntSetFont(oldFont);
    if (FeatureSet.OS35) {
        WinPopDrawState();
    }
}

static Boolean
_GridHandleEvent(GridType* grid, EventType* e)
{
    UInt16 screenRow, col;
    Coord x, y;
    Boolean ret = false;

    switch (e->eType) {
    case penDownEvent:
        x = e->screenX;
        y = e->screenY;

#ifdef CONFIG_SONY
	    if( FeatureSet.Sony && SonyLibRefNum) {
			x *= 2;
			y *= 2;
		} else 
#endif
        if (FeatureSet.HiRes) {
            x = WinScaleCoord(x, true);
            y = WinScaleCoord(y, true);
        }

		WinDisplayToWindowPt(&x, &y);

        if (! RctPtInRectangle(x, y, &(grid->bounds)) ||
                y > grid->bounds.topLeft.y + grid->headerExtent + grid->numScreenRows * grid->lineHeight) {
            ret = false;
	    break;
        }

       switch (PtToPosition(grid, &screenRow, &col, x, y)) {
        case gridPositionNone:
            break;

        case gridPositionHeaderCell:
            /* Only continue if the header is configured as selectable. */
            if (! grid->header_selectable)
                break;

            SelectHeaderCell(grid, col);
        break;

        case gridPositionHeaderColumnSep:
            /* No action right now. Maybe resize the columns by dragging? */
        break;

        case gridPositionColumnSep:
            /* No action right now. Maybe resize the columns by dragging? */
        break;

        case gridPositionCell:
            /* If this row is being used, then bail out. */
            if (! grid->rows[TopRowIndex(grid) + screenRow].usable)
                break;

            switch (grid->model->GetCellType(grid->model_data,
                             grid->rows[TopRowIndex(grid) + screenRow].modelIndex,
                             col)) {
            case gridCellTypeNote:
                SelectCell(grid, screenRow, col);
            break;
            case gridCellTypeBoolean:
                if (grid->model->IsColumnEditable(grid->model_data, col)) {
                    SelectCell_Boolean(grid, screenRow, col, e);
                    break;
                }

                /* fall-through if the column is not editable */

            default:
                SelectCell(grid, screenRow, col);
            break;
            }
        break;

        }
        ret = true;
    break;
    default:
    break;

    }

    return ret;
}

/**
 * LoadCache:
 *  
 *  ensure that the cache of row information is up to date when
 *  the topRow of the grid changes.
 */
static void 
LoadCache(GridType *grid) {
	UInt16 i, nextIndex;


	/*
	 * first check to see if all the rows that will be visible are already
	 * in the cache.  If so, on loading needs to be done. 
	 */

	if ( grid->topRow >= grid->rowCacheOffset &&
		 grid->rowCacheOffset + grid->numCacheRows >= grid->topRow + grid->numScreenRows ) {
		return;

	}

	if ( grid->topRow < grid->rowCacheOffset ) {
		/* Page Up:  This will only happen if we're in VLDB mode... */
		nextIndex = grid->rows[0].modelIndex;

		while( grid->rowCacheOffset > grid->topRow ) {
			grid->model->Seek(grid->model_data, &nextIndex, 1, dmSeekBackward);
			grid->rowCacheOffset--;
		}

	}
	else if ( grid->topRow >= grid->rowCacheOffset + grid->numCacheRows ) {
		/* Page Down:  This will only happen if we're in VLDB mode... */
		nextIndex = grid->rows[grid->numCacheRows - 1].modelIndex;

		grid->rowCacheOffset += grid->numCacheRows - 1;
		while( grid->rowCacheOffset < grid->topRow ) {
			grid->model->Seek(grid->model_data, &nextIndex, 1, dmSeekForward);
			grid->rowCacheOffset++;
		}


	}	
	else {
		// page up or down, but to some place that's already loaded
		// this could be optimized with a bitblp

		nextIndex = grid->rows[TopRowIndex(grid)].modelIndex;

		grid->rowCacheOffset = grid->topRow;

	}
	for( i = 0; i < grid->numCacheRows; i++ ) {
		grid->rows[i].modelIndex = nextIndex;
		grid->model->Seek(grid->model_data, &(grid->rows[i].modelIndex), 0, dmSeekForward);
		grid->rows[i].height = grid->lineHeight;
		grid->rows[i].usable = false;
		grid->rows[i].valid = false;
		grid->rows[i].highlighted = false;
		nextIndex = grid->rows[i].modelIndex + 1;
	}
}

static Boolean
_GridScrollVertical(GridType* grid, WinDirectionType direction,
            UInt16 increment)
{
    int i;
    UInt16 distance;
    RectangleType r, dummy;
    WinHandle winH;

	if ((direction == winDown) && (grid->topRow + grid->numScreenRows == grid->numModelRows))
		return false;
	if ((direction == winUp) && (!grid->topRow))
		return false;

    /* If the increment is greater than or equal to the number of
     * visible rows, then it is simpler to invalidate every row and
     * redraw everthing. GridDraw() outputs everything to the screen
     * at once so there will be no flicker.
     */

    if (increment >= grid->numScreenRows) {
        if (direction == winDown) {
            grid->topRow += increment;
        } else {
            if (grid->topRow > increment)
                grid->topRow -= increment;
            else
                grid->topRow = 0;
        }

        /* Invalidate all of the rows onscreen. */
		if (increment < grid->numCacheRows )
			InvalidateGrid(grid);

		LoadCache( grid );
        /* Redraw the screen and return to caller. */
        _GridDraw(grid);
        return true;
    }

    /* Otherwise, we will move the rows that will stay on screen up
     * the offscreen buffer and then force a redraw on the new rows
     * that have been opened up.
     */

    /* Determine what the new topmost row will be. */
    if (direction == winDown) {

        if (grid->topRow + increment < grid->numModelRows - grid->numScreenRows)
            grid->topRow += increment;
        else {
            increment = grid->numModelRows - grid->topRow - grid->numScreenRows;
            grid->topRow = grid->numModelRows - grid->numScreenRows;
        }
		if ( grid->topRow + grid->numScreenRows >= grid->rowCacheOffset + grid->numCacheRows ) {

			LoadCache(grid);
			
		}
        for (i = grid->numScreenRows-increment; i < grid->numScreenRows; ++i) {
            grid->rows[TopRowIndex(grid) + i].usable = false;
            grid->rows[TopRowIndex(grid) + i].valid = false;
        }
    } else {
        if (grid->topRow > increment)
            grid->topRow -= increment;
        else {
            increment = grid->topRow;
            grid->topRow = 0;
        }
		if ( grid->topRow < grid->rowCacheOffset ) {
	
			// reload cache
			LoadCache(grid);
	
		}
       for (i = 0; i < increment; ++i) {
            grid->rows[TopRowIndex(grid) + i].usable = false;
            grid->rows[TopRowIndex(grid) + i].valid = false;
        }
    }
    /* Switch to the offscreen window. */
    winH = WinGetDrawWindow();
    WinSetDrawWindow(grid->win);

    /* Move the rows within the offscreen window. */
    r.topLeft.x = 0;
    r.topLeft.y = grid->headerExtent;
    r.extent.x = grid->bounds.extent.x;
    r.extent.y = grid->numScreenRows * grid->lineHeight;
    distance = 0;
    if (direction == winDown) {
        for (i = 0; i < increment; ++i) {
            distance += grid->rows[TopRowIndex(grid) + i].height;
        }
#ifdef CONFIG_SONY
        if (FeatureSet.Sony && SonyLibRefNum) {
            HRWinScrollRectangle(SonyLibRefNum, &r, winUp, distance, &dummy);
        } else 
#endif
        WinScrollRectangle(&r, winUp, distance, &dummy);
    } else {
        for (i = grid->numScreenRows - increment; i < grid->numScreenRows; ++i) {
            distance += grid->rows[TopRowIndex(grid) + i].height;
        }
#ifdef CONFIG_SONY
        if (FeatureSet.Sony && SonyLibRefNum) {
            HRWinScrollRectangle(SonyLibRefNum, &r, winDown, distance, &dummy);
        } else 
#endif
        WinScrollRectangle(&r, winDown, distance, &dummy);
    }
    WinEraseRectangle(&dummy, 0);

	/**
	   We cannot just draw the column seperator for the rectangle we just
	   erased because GridDraw may erase that area again when it draws
	   the new data.  Instead we should do this...
	*/
	grid->valid &= ~Valid_Col_sep;

    /* Switch back to the original draw window. */
    WinSetDrawWindow(winH);

    /* Draw the rows that have been displaced. */
    _GridDraw(grid);

	return true;
}

static Boolean
_GridScrollHorizontal(GridType* grid, WinDirectionType direction,
              UInt16 increment)
{
	Boolean ret = false;

    if (direction == winRight) {
        if (grid->leftColumn + increment < grid->numCols)
		{
            grid->leftColumn += increment;
			ret = true;
		}
    } else {
		if (grid->leftColumn)
		{
			if (increment > grid->leftColumn)
				grid->leftColumn = 0;
			else
				grid->leftColumn -= increment;
			ret = true;
		}
    }

	if (ret)
	{
		grid->rightColumn = grid->leftColumn;
		
		FillColumnInfo(grid);
		InvalidateGrid(grid);
	}

	return ret;
}

static void
_GridSetHeaderLook(GridType* grid, Boolean display_header,
               Boolean display_header_sep,
               Boolean header_selectable, FontID header_font)
{
    if (display_header_sep)
        grid->display |= Display_Header_sep;
    else
        grid->display &= ~(Display_Header_sep);

    grid->header_selectable = header_selectable;
    grid->header_font = header_font;

    if (display_header) {
        grid->display |= Display_Header;

        grid->headerExtent = FntLineHeight() + 1;

        if (grid->display & Display_Header_sep)
            grid->headerExtent += 1;
    } else {
        grid->display &= ~(Display_Header);
        grid->headerExtent = 1;
    }
}

static void
_GridSetColumnLook(GridType* grid, Boolean display_colsep)
{
    if (display_colsep) {
        grid->display |= Display_Col_sep;
        grid->columnGutterExtent = 1;
    } else {
        grid->display &= ~(Display_Col_sep);
        grid->columnGutterExtent = 0;
    }

    grid->rightColumn = grid->leftColumn;
}

static UInt16
_GridGetNumberOfRows(GridType* grid)
{
    return grid->numScreenRows;
}

static void
_GridSetNumberOfRows(GridType* grid, UInt16 number)
{
}

static void 
_GridHighlightIndex(GridType* grid, UInt16 index)
{
    UInt16 i;
    /* Now scan the list of visible indexes and force a redraw for the
     * index if it is present.
     */
    for (i = 0; i < grid->numCacheRows; ++i) {
        if (grid->rows[i].valid && grid->rows[i].modelIndex == index) {
            grid->rows[i].valid = false;
            grid->rows[i].highlighted = true;
        }
    }
}

static void
_GridUnhighlightAll(GridType* grid)
{
    UInt16 i;

    /* Force a redraw for any rows that are highlighted. */
    for (i = 0; i < grid->numCacheRows; ++i) {
        if (grid->rows[i].valid && grid->rows[i].highlighted) {
            grid->rows[i].valid = false;
            grid->rows[i].highlighted = false;
        }
    }
}

static UInt16
_GridGetTopIndex(const GridType* grid)
{
    return grid->rows[TopRowIndex(grid)].modelIndex;
}

static void
_GridSetTopIndex(GridType* grid, UInt16 index)
{
    UInt16 row, mi;

    if (grid->rows[TopRowIndex(grid)].modelIndex == index) {
        InvalidateGrid(grid);
        return;
    }

    if (index > grid->lastModelIndex) {
        grid->topRow = grid->numModelRows - 1;
    } else if (index < grid->rows[TopRowIndex(grid)].modelIndex) {
		if ( index >= grid->rows[0].modelIndex ) {

			for (row = 0; row < TopRowIndex(grid); ++row) {
				if (grid->rows[row].modelIndex == index) {
					grid->topRow = grid->rowCacheOffset + row;
					break;
				}
			}
		}
		else {
			// need to seek

			mi = grid->rows[0].modelIndex;
			while( index < mi && grid->topRow > 0) {
				grid->model->Seek(grid->model_data, &mi, 1, dmSeekBackward );
				grid->topRow--;
			}

		}
    } else if ( index > grid->rows[grid->numCacheRows-1].modelIndex ) {
		// need to seek forward
		mi = grid->rows[grid->numCacheRows-1].modelIndex;
		while( index > mi ) {
			if ( !grid->model->Seek(grid->model_data, &mi, 1, dmSeekForward ) ) break;
			grid->topRow++;
		}		
		
	
    } else {
        for (row = TopRowIndex(grid); row < grid->numCacheRows; ++row) {
            if (grid->rows[row].modelIndex == index) {
				grid->topRow = grid->rowCacheOffset + row;
                break;
            }
        }
    }

    InvalidateGrid(grid);
	LoadCache(grid);
}

#ifdef CONFIG_JOGDIAL
static void
_GridSelectRowIndex(GridType* grid, UInt16 index)
{
    UInt16 i;

//    if (Gridp->screenRowSelected != NoneScreenRowSelected)
//        return;

	if (index == dmMaxRecordIndex)
	{
		grid->screenRowSelected = NoneScreenRowSelected;
		return;
	}

    for (i = 0; i < grid->numScreenRows; i++) {
        if (grid->rows[TopRowIndex(grid) + i].modelIndex == index) {
            grid->screenRowSelected = grid->topRow + i;
            return;
        }
    }
	/*
	  This won't work with the VLDB code...

    if (grid->rows[grid->topRow].modelIndex > index) {
        for (i = 0; i < grid->topRow; i++) {
            if (grid->rows[i].modelIndex == index) {
                grid->topRow = i;
                grid->screenRowSelected = i;
                return;
            }
        }
    }
    for (i = grid->numScreenRows; i < grid->numCacheRows; i++) {
        if (grid->rows[i].modelIndex == index) {
            grid->topRow = i;
            grid->screenRowSelected = i;
            return;
        }
    }
	*/
}

static Boolean
_GridMoveSelectedLine(GridType* grid, WinDirectionType direction)
{
	Boolean ret = false;
    Int16 row = grid->screenRowSelected;

	if ( row >= grid->rowCacheOffset &&
		 ( row - grid->rowCacheOffset ) < grid->numCacheRows ) {

		grid->rows[row - grid->rowCacheOffset].valid = false;

	}

    switch (direction) {
    case winDown:
        row++;
    break;
    case winUp:
        row--;
    break;
    default:
        // do nothing
    }
    if ( row >= grid->rowCacheOffset  && 
		 (row - grid->rowCacheOffset ) < grid->numCacheRows) {

        grid->rows[row - grid->rowCacheOffset].valid = false;
        grid->screenRowSelected = row;
		ret = true;
    }

    _GridDraw(grid);

	return ret;
}

#endif

GridOpsType StandardGridOps =
{
    GridInit:               _GridInit,
    GridFree:               _GridFree,
    GridChange:             _GridChange,
    GridDraw:               _GridDraw,
    GridHandleEvent:        _GridHandleEvent,
    GridSetTopIndex:        _GridSetTopIndex,
    GridGetTopIndex:        _GridGetTopIndex,
    GridGetNumberOfRecordsInPage: _GridGetNumberOfRows,
    GridSetNumberOfRecordsInPage: _GridSetNumberOfRows,
    GridScrollVertical:     _GridScrollVertical,
    GridScrollHorizontal:   _GridScrollHorizontal,
    GridSelectRecordIndex:     _GridSelectRowIndex,
    GridMoveSelectedLine:   _GridMoveSelectedLine,
    GridHighlightIndex:     _GridHighlightIndex,
    GridUnhighlightAll:     _GridUnhighlightAll,
    GridSetHeaderLook:      _GridSetHeaderLook,
    GridSetColumnLook:      _GridSetColumnLook,
    GridGetFieldBounds:      _GridGetCellBounds,
    GridGetRecordBounds:       _GridGetRowBounds,
    GridGetHeaderBounds:    _GridGetHeaderBounds
};

GridType StandardGrid;

GridType *
Grid_New()
{
#ifdef CONFIG_PLUGINS
    GridType * grid;

    if ((grid = Plug_GetGrid()))
        return grid;
#endif
    StandardGrid.ops = &StandardGridOps;
    return &StandardGrid;
}

void
Grid_Free(GridType *grid)
{
}

#if 0
static void _GridMakeIndexVisible(GridType* grid, UInt16 index) GRIDSECT;

static IndexedColorType _GridGetColor(const GridType*,
				     GridUIColorTableEntries) GRIDSECT;

static void _GridSetColor(GridType*, GridUIColorTableEntries,
			IndexedColorType) GRIDSECT;

static void
_GridMakeIndexVisible(GridType* grid, UInt16 index)
{
    UInt16 row;

    /* Check to see if the row is already visible. If so, return. */
    for (row = 0; row < grid->numScreenRows; ++row) {
        if (grid->rows[TopRowIndex(grid) + row].usable
                && grid->rows[TopRowIndex(grid) + row].modelIndex == index)
            return;
    }

    _GridSetTopIndex(grid, index);
    for (row = 0; row < grid->numScreenRows; ++row) {
        grid->rows[TopRowIndex(grid) + row].usable = false;
        grid->rows[TopRowIndex(grid) + row].valid = false;
    }
}

#ifdef CONFIG_COLOR
static IndexedColorType
_GridGetColor(const GridType* grid, GridUIColorTableEntries index)
{
    return grid->colors[index];
}

static void
_GridSetColor(GridType* grid, GridUIColorTableEntries index,
         IndexedColorType color)
{
    grid->colors[index] = color;
}
#endif

#endif
