/*
 * DB: Feature Names Editor
 * Copyright (C) 2000 by Tom Dyas (tdyas@users.sourceforge.net)
 * Copyright (C) 2002 by Scott Wallace (swallace@eecs.umich.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * This file contains the GUI code which allows the editing of
 * features (e.g. list views, filters, etc) appearing in a
 * database. Actual editing of a feature is left to more specialized
 * forms and handlers.  
 */

#include <PalmOS.h>

#include "db.h"
#ifdef CONFIG_HANDERA
#include <Vga.h>
#endif

#include "util.h"
#include "enum.h"
#include "names.h"
#include "localsearch.h"



#define GetObjectPtr(f,i) FrmGetObjectPtr((f),FrmGetObjectIndex((f),(i)))

static void AppendNewFeature(DataSource * source, FeatureType f) SECT_UI;
static void UpdateButtons(FormType* form) SECT_UI;

FeatureType FeatureEditor_Feature;
UInt16 FeatureEditor_FeatureNum;

/* Return an array of strings containing the names of all the list views. */
void
GetFeatureNames(char*** names_ptr, UInt16* count_ptr, FeatureType f, char *first)
{
    char ** names = NULL;
    UInt16 i, count, others;

    count = CurrentSource->ops.GetNumFeatures(CurrentSource, f);
	if ( first != NULL ) others = 1;
	else others = 0;

    if (count || others ) {
        names = AllocateMemPtr( (count+others+1) * sizeof(char *));
		if ( others ) {
			names[0] = AllocateMemPtr( 1 + StrLen( first ) );
			StrCopy( names[0], first );
		}
		for ( i = others; i < count+others; i++ ) {
            DataSourceFeature* ftr = CurrentSource->ops.GetFeature(CurrentSource, f, i-others);
            names[i] = AllocateMemPtr(1 + StrLen(ftr->name));
            StrCopy(names[i], ftr->name);
            CurrentSource->ops.PutFeature(CurrentSource, f, i-others, ftr);
        }
		names[count+others] = NULL;
    }

    *names_ptr = names;
    *count_ptr = count+others;
}

/* Free the memory associated with the feature names list. */
void
PutFeatureNames(char ** names)
{
    UInt16 i;

	if ( names == NULL ) return;

    for (i = 0; names[i]; i++) {
        DeallocateMemPtr(names[i]);
    }
    if (names)
        DeallocateMemPtr(names);
}


/* Append a new (default, or blank) feature to the data source. */
static void
AppendNewFeature(DataSource * source, FeatureType f)
{
	UInt16 num;
	char *name;

	switch ( f ) {
	case featureListView:
	{

		DataSourceFeature_ListView view;
		DataSourceViewColumn cols[source->schema.numFields];
		UInt16 i;

		name = GetStringResource(stringID_AllFields);

		MemSet(&view, sizeof(view), 0);
		StrCopy(view.name, name);
		PutStringResource(name);
		view.numCols = source->schema.numFields;
		view.cols = &cols[0];
		for (i = 0; i < view.numCols; ++i) {
			view.cols[i].fieldNum = i;
			view.cols[i].width = 80;
		}

		num = source->ops.GetNumFeatures(source, featureListView);
		source->ops.StoreFeature(source, featureListView, num, &view);
	}
	break;
	#ifdef CONFIG_GLB_SCRIPTS
	case featureGlobalScript:
	{
		DataSourceFeature_GlobalScript gsd;
		
		name = GetStringResource(stringID_GlobalScript);
		
		MemSet(&gsd, sizeof(DataSourceFeature_GlobalScript), 0);
		StrCopy(gsd.name, name);
		PutStringResource(name);
		num = source->ops.GetNumFeatures(source, featureGlobalScript);
		source->ops.StoreFeature(source, featureGlobalScript, num, &gsd);

	}
	break;
#endif
	case featureFilter:
	{
		DataSourceFeature_Filter filt;
		SimpleFilterData filt_data;

		name = GetStringResource(stringID_NewFilter);
		
		MemSet(&filt, sizeof(DataSourceFeature_Filter), 0);
		StrCopy(filt.name, name);
		PutStringResource(name);

		filt.data.simple = &filt_data;
		filt_data.matchString[0] = '\0';
		num = source->ops.GetNumFeatures(source, featureFilter);
		source->ops.StoreFeature(source, featureFilter, num, &filt);

	}
	break;
	
	default:
	}
}


/* Show/hide actions buttons based on the data source's capabilities. */
static void
UpdateButtons(FormType* form)
{
	UInt8 capable;
	
	capable = CurrentSource->ops.SupportsFeature(CurrentSource, FeatureEditor_Feature);

	
#define ShowObject(f,id) FrmShowObject((f),FrmGetObjectIndex((f),(id)))
#define HideObject(f,id) FrmHideObject((f),FrmGetObjectIndex((f),(id)))


	if ( FeatureEditor_Feature == featureGlobalScript )
		ShowObject(form, ctlID_NamesEditor_ExeButton);
	else
		HideObject(form, ctlID_NamesEditor_ExeButton);

    if (capable & FEATURE_SUPPORT_INSERT)
        ShowObject(form, ctlID_NamesEditor_NewButton);
    else
        HideObject(form, ctlID_NamesEditor_NewButton);

    if (capable & FEATURE_SUPPORT_DELETE)
        ShowObject(form, ctlID_NamesEditor_DeleteButton);
    else
        HideObject(form, ctlID_NamesEditor_DeleteButton);

#undef ShowObject
#undef HideObject
}

Boolean
NamesEditorHandleEvent(EventType* event)
{
    static char ** names;
    static char* label;

    FormType* form;
    ControlType* ctl;
    ListType* list;
    UInt16 count;
	UInt8 support;

    switch (event->eType) {
    case frmOpenEvent:
        form = FrmGetActiveForm();
        names = NULL;
        label = NULL;
#ifdef CONFIG_HANDERA
        if (FeatureSet.Handera) {
            VgaFormModify(form, vgaFormModify160To240);
            if (!UtilCheckSizeForm(form, 0)) {
                FrmAlert(alertID_FormTooBig);
                FrmGotoForm(formID_ListView);
                return true;
            }
        }
#endif
		switch ( FeatureEditor_Feature ) {
		case featureListView:
			FrmSetTitle( form, "DB: List Views" );
			break;
#ifdef CONFIG_GLB_SCRIPTS
		case featureGlobalScript:
			FrmSetTitle( form, "DB: Global Scripts" );
			break;
#endif			
		default:
			FrmSetTitle( form, "DB: Filters" );
			break;
		}
        GetFeatureNames(&names, &count, FeatureEditor_Feature, NULL);
        list = GetObjectPtr(form, ctlID_NamesEditor_List);
        LstSetListChoices(list, names, count);
        LstSetSelection(list, noListSelection);

		
		support = CurrentSource->ops.SupportsFeature( CurrentSource,
													  FeatureEditor_Feature );
		
        if ( support & FEATURE_SUPPORT_EDIT) {
            label = CopyStringResource(stringID_Edit);
        } else {
            label = CopyStringResource(stringID_View);
        }
        ctl = GetObjectPtr(form, ctlID_NamesEditor_EditButton);
        CtlSetLabel(ctl, label);

        UpdateButtons(form);
        FrmDrawForm(form);
        return true;
		
	case frmUpdateEvent:
		if ( event->data.frmUpdate.updateCode == updateChangeForm ) {
		form = FrmGetActiveForm();


		PutFeatureNames(names);					
		GetFeatureNames(&names, &count, FeatureEditor_Feature, NULL);
		list = GetObjectPtr(form, ctlID_NamesEditor_List);
		LstSetListChoices(list, names, count);
		LstSetSelection(list, noListSelection);
		LstDrawList( list );
		FrmDrawForm( form );
		return true;
		}
		return false;

    case frmCloseEvent:
		PutFeatureNames(names);

        if (label)
            DeallocateMemPtr(label);
        break;

    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
        case ctlID_NamesEditor_DoneButton:
            FrmGotoForm(formID_ListView);
            return true;

        case ctlID_NamesEditor_EditButton:
            form = FrmGetActiveForm();
            list = GetObjectPtr(form, ctlID_NamesEditor_List);
            count = LstGetSelection(list);
            if (count != noListSelection) {

                FeatureEditor_FeatureNum = count;
				switch (FeatureEditor_Feature) {
				case featureListView:
					FrmGotoForm(formID_ListViewEditor);
					break;
#ifdef CONFIG_GLB_SCRIPTS
				case featureGlobalScript:
					FrmPopupForm( formID_GlobalScriptEdit );
					break;
#endif
				case featureFilter:
					FindInfo.forFilterDefinition = FeatureEditor_FeatureNum;
					FrmPopupForm(formID_FindDialog);
					break;
				default:
					break;
				}
            }
            return true;

        case ctlID_NamesEditor_NewButton:
            PutFeatureNames(names);
            AppendNewFeature(CurrentSource, FeatureEditor_Feature);
            GetFeatureNames(&names, &count, FeatureEditor_Feature, NULL);
            form = FrmGetActiveForm();
            list = GetObjectPtr(form, ctlID_NamesEditor_List);
            LstSetListChoices(list, names, count);
            LstSetSelection(list, noListSelection);
            LstEraseList(list);
            LstDrawList(list);
            UpdateButtons(form);
            return true;
			
        case ctlID_NamesEditor_DeleteButton:
            form = FrmGetActiveForm();
            list = GetObjectPtr(form, ctlID_NamesEditor_List);
            count = LstGetSelection(list);
            if (count != noListSelection) {
                PutFeatureNames(names);
                CurrentSource->ops.DeleteFeature(CurrentSource, FeatureEditor_Feature, count);
                GetFeatureNames(&names, &count, FeatureEditor_Feature, NULL);
                form = FrmGetActiveForm();
                list = GetObjectPtr(form, ctlID_NamesEditor_List);
                LstSetListChoices(list, names, count);
                LstSetSelection(list, noListSelection);
                LstEraseList(list);
                LstDrawList(list);
                UpdateButtons(form);
            }
            return true;

        case ctlID_NamesEditor_ExeButton:
            list = GetObjectPtr(FrmGetActiveForm(), ctlID_NamesEditor_List);
            count = LstGetSelection(list);
            if (count != noListSelection) {
				DataSourceFeature_GlobalScript *gsd;
				ScriptContext sc;
				UInt8 *bc;
				CalculatedValue cv;

                gsd = CurrentSource->ops.GetFeature(CurrentSource, featureGlobalScript, count);
				sc.context = GENERAL_GLOBAL_SCRIPT_CONTEXT;
				sc.src = CurrentSource;
				sc.flags = ScriptContext_DataAccess_NFA;
				
				if ( gsd->c.bytecode ) {
					bc = MemHandleLock( gsd->c.bytecode );
					CurrentSource->ops.ExecuteScript( &sc, bc, &cv );
					MemPtrUnlock( bc );
				}
				CurrentSource->ops.PutFeature(CurrentSource, featureGlobalScript, count, gsd );
			}
            return true;

        }
        break;

    default:
        break;
    }

    return false;
}
